/* monomial hash table */

/* Maple seq of variables */
static ALGEB  vlist = NULL;
static INT32  nvars = 0;

/* monomial exponents block */
static INT32 *expon = NULL;
static INT32  esize = 0;
static INT32  eload = 1;

/* monomials to expon index */
static INT32 *table = NULL;
static INT32  tsize = 0;

/* random integers for hash */
static INT32 *trand = NULL;

#define HASH_DEG (nvars+0)	/* stores total degree */
#define HASH_VAL (nvars+1)	/* original hash value */
#define HASH_DIV (nvars+2)	/* cache for divisions */
#define HASH_IND (nvars+3)	/* index into an array */
#define HASH_LEN (nvars+4)	/* length of monomials */


static void table_init(ALGEB v)
{
	INT32 i;

	vlist = v;
	nvars = (INT32)LENGTH(v)-1;

	/* hash table */
	tsize = TABLESIZE;
	trand = malloc(nvars*sizeof(INT32));
	table = malloc(tsize*sizeof(INT32));

	/* exponent vectors */
	esize = HASH_LEN*tsize/2;
	eload = HASH_LEN;
	expon = malloc(esize*sizeof(INT32));
	for (i=0; i < HASH_LEN; i++) expon[i] = 0;

	/* no zeroes allowed in hash vector */
	for (i=0; i < nvars; i++) trand[i] = randgen() | 1;
	for (i=0; i < tsize; i++) table[i] = 0;

}


static void table_free()
{
	if (table) free(table);
	if (trand) free(trand);
	if (expon) free(expon);
	tsize = esize = eload = nvars = 0;
	table = NULL;
	trand = NULL;
	expon = NULL;
}


static void table_enlarge()
{
	INT32 h, i, j, k, *e;

	esize = 2*esize;
	expon = realloc(expon, esize*sizeof(INT32));

	tsize = 2*tsize;
	table = realloc(table, tsize*sizeof(INT32));
	for (i=0; i < tsize; i++) table[i] = 0;

	/* re-insert all elements in block */
	for (i=HASH_LEN; i < eload; i+=HASH_LEN) {
		e = expon + i;
		h = e[HASH_VAL];

		/* quadratic probe insert */
		for (k=h, j=0; j < tsize; j++) {
			k = (k + j) & (tsize-1);
			if (table[k]) continue;
			table[k] = i;
			break;
		}
	}
}


static INT32 table_insert(INT32 *ev)
{
	INT32 h, i, j, k, *e, m, deg;

	/* compute hash value */
	for (h=0, i=0; i < nvars; i++) {
		h += trand[i] * ev[i];
	}

	/* quadratic probe lookup */
	for (k=h, j=0; j < tsize; j++) {
		k = (k + j) & (tsize-1);
		if (!table[k]) break;
		e = expon + table[k];
		if (e[HASH_VAL] != h) continue;
		for (i=0; i < nvars; i++) {
			if (e[i] != ev[i]) break;
		}
		if (i==nvars) return table[k];
	}

	/* create new element */
	m = eload;
	e = expon + m;
	for (deg=0, i=0; i < nvars; i++) {
		e[i] = ev[i];
		deg += ev[i];
	}
	e[HASH_DEG] = deg;
	e[HASH_VAL] = h;
	e[HASH_DIV] = 0;
	e[HASH_IND] = 0;
	table[k] = m;

	eload += HASH_LEN;
	if (eload >= esize) {
		table_enlarge();
	}
	return m;
}


static INT32 table_insert_prod(INT32 *f, INT32 *g)
{
	INT32 h, i, j, k, *e, m, deg;

	/* compute hash value */
	h = f[HASH_VAL] + g[HASH_VAL];

	/* quadratic probe lookup */
	for (k=h, j=0; j < tsize; j++) {
		k = (k + j) & (tsize-1);
		if (!table[k]) break;
		e = expon + table[k];
		if (e[HASH_VAL] != h) continue;
		for (i=0; i < nvars; i++) {
			if (e[i] != f[i] + g[i]) break;
		}
		if (i==nvars) return table[k];
	}

	/* create new element */
	m = eload;
	e = expon + m;
	for (i=0; i <= HASH_DEG; i++) {
		e[i] = f[i] + g[i];
	}
	e[HASH_VAL] = h;
	e[HASH_DIV] = 0;
	e[HASH_IND] = 0;
	table[k] = m;

	eload += HASH_LEN;
	if (eload >= esize) {
		table_enlarge();
	}
	return m;
}
