/* sparse linear algebra */

/* matrix format: [ length | row1 | row2 | ... ] */
/*    row format: [ length | deg1 | cof1 | deg2 | cof2 | ... ] */


/* free matrix structure */
static void matrix_free(INT32 **a)
{
	M_INT i, l;
	l = (M_INT)(a[0]);
	for (i=1; i < l; i++) {
		free(a[i]);
	}
	free(a);
}


/* invert leading coefficient mod p */
static void matrix_normal_row(INT32 *a, INT32 p)
{
	INT32 l,i,c;
	INT64 t;
	c = modinv32(a[2], p);
	for (l=a[0], i=2; i < l; i+=2) {
		t = ((INT64)a[i] * c) % p;
		// t = (t < 0) ? t+p : t;
		t += (t >> 63) & p;
		a[i] = (INT32)t;
	}
}


/* reduce row in buffer with pivots */
static INT32 * matrix_reduce_row(INT32 *a, INT32 **pivots, INT64 *buffer, INT32 p)
{
	INT32 lt, i, j, k, l;
	INT64 p2, c, t0, t1;

	/* write row into buffer */
	for (l=a[0], i=1; i < l; i+=2) {
		buffer[a[i]] = (INT64)a[i+1];
	}
	lt = a[1];
	free(a);

	p2 = (INT64)p*p;

	reduce:
	/* k = nb. of non-zero elements */
	/* lt = index of first non-zero */
	for (k=0, i=lt, lt=-1, k=0; i >= 0; i--) {
		c = buffer[i];
		if (c) c %= p;
		buffer[i] = c;
		if (!c) continue;

		a = pivots[i];
		if (!a) {
			if (!k) lt = i;
			k++;
			continue;
		}
#if 0
		buffer[i] = 0;
		for (j=3; j < a[0]; j+=2) {
			t0 = buffer[a[j]];
			t0 -= c * a[j+1];
			// t0 = (t0 < 0) ? t0+p2 : t0;
			t0 += (t0 >> 63) & p2;
			buffer[a[j]] = t0;
		}
#else
		/* 2x unroll */
		j = (a[0]-1)/2;
		j = (j & 1) ? 3 : 1;
		for (; j < a[0]; j+=4) {
			t0 = buffer[a[j+0]];
			t1 = buffer[a[j+2]];
			t0 -= c * a[j+1];
			t1 -= c * a[j+3];
			t0 += (t0 >> 63) & p2;
			t1 += (t1 >> 63) & p2;
			buffer[a[j+0]] = t0;
			buffer[a[j+2]] = t1;
		}
		buffer[i] = 0;
#endif
		/* check for interim pivot */
		if (k && pivots[lt]) goto reduce;
	}
	if (k==0) return NULL;

	/* check for interim pivot */
	if (k && pivots[lt]) goto reduce;

	/* extract row from buffer */
	a = malloc((2*k+1)*sizeof(INT32));
	a[0] = 2*k+1;
	for (j=1, i=lt; i >= 0; i--) {
		if (!buffer[i]) continue;
		a[j++] = i;
		a[j++] = (INT32)buffer[i];
		buffer[i] = 0;
	}
	if (a[2] != 1) matrix_normal_row(a,p);
	return a;
}


/* reduce to a new pivot and insert */
static void matrix_insert_row(INT32 *a, INT32 **pivots, INT64 *buffer, INT32 m, INT32 p)
{
	M_INT t, i;

	t = __cilkrts_get_worker_number();
	buffer = buffer + t*m;
	do {
		a = matrix_reduce_row(a, pivots, buffer, p);
		if (!a) return;
		i = cas( (void *)(&pivots[a[1]]), 0, I(a));
	} while (i);
}


/* comparison used for qsort (ascending) */
static int matrix_cmp(const void *a, const void *b)
{
	M_INT d, e;
	d = ((INT32 **)a)[0][1];
	e = ((INT32 **)b)[0][1];
	if (d > e) return  1;
	if (d < e) return -1;
	d = ((INT32 **)a)[0][0];
	e = ((INT32 **)b)[0][0];
	if (d > e) return  1;
	if (d < e) return -1;
	return 0;
}


/* ascending by pivot */
static INT32 ** matrix_sort(INT32 **A)
{
	qsort( A+1, I(A[0])-1, sizeof(M_INT), &matrix_cmp );
	return A;
}


/* row reduce sorted matrix: flag = fast */
static INT32 ** matrix_ref(INT32 **A, INT32 p, INT32 flag)
{
	INT64 *buffer;
	INT32 *a, **pivots;
	INT32 n, m, i, j;
	M_INT k, t;

	double c0, c1, t0, t1;
	c0 = cputime(), t0 = realtime();

	matrix_sort(A);
	n = I(A[0])-1;
	m = A[n][1]+1;

	t = __cilkrts_get_nworkers();
	for (k=0, i=1; i <= n; i++) {
		k += (A[i][0]-1)/2;
	}
	DEBUG(MATDB, "matrix:   %d x %d, %5.1f per row, %5.1f MB\n", n, m, 1.0*k/n, (2.0*k+n)*sizeof(INT32)/1024/1024);

	/* initialize scratch space */
	pivots = malloc(m*sizeof(M_INT));
	for (i=0; i < m; i++) pivots[i] = NULL;
	buffer = malloc(t*m*sizeof(INT64));
	for (i=0; i < t*m; i++) buffer[i] = 0;

	/* initialize pivots */
	for (i=j=1; i < I(A[0]); i++) {
		if (flag && !pivots[A[i][1]])
			pivots[A[i][1]] = A[i];
		else
			A[j++] = A[i];
	}
	A[0] = (INT32 *)I(j);

	/* reduce rows in order */
	for (i=1; i < I(A[0]); i++) {
		if (flag) {
			cilk_spawn matrix_insert_row(A[i], pivots, buffer, m, p);
		}
		else {
			matrix_insert_row(A[i], pivots, buffer, m, p);
		}
	}
	cilk_sync;

	/* form ref matrix from pivots */
	for (i=1, j=0, k=0; j < m; j++) {
		if (!pivots[j]) continue;
		A[i++] = pivots[j];
		k += (pivots[j][0]-1)/2;
	}
	A[0] = (INT32 *)I(i);
	A = realloc(A, i*sizeof(M_INT));
	free(pivots);
	free(buffer);

	c1 = cputime(), t1 = realtime();
	DEBUG(MATDB, "matrix:   %d x %d, %5.1f per row, %5.1f MB, %d zero\n", i-1, m, 1.0*k/(i-1), (2.0*k+n)*sizeof(INT32)/1024/1024, n-i+1);
	DEBUG(MATDB, "matrix:   %5.3f sec   %5.3f cpu   %3.f%%\n", t1-t0, c1-c0, 100*(c1-c0)/(t1-t0));
	matrix_time += t1-t0;
	matrix_cput += c1-c0;

	return A;
}
